from typing import Iterable, Set

from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet


def list_codepends_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    transitive: bool = True,
    include_selected: bool = True,
) -> Iterable[str]:
    result: Set[str] = set(addons_selection) if include_selected else set()
    codeps = direct_codependencies(addons_selection, addons_set, result)
    result |= codeps
    while transitive and codeps:
        codeps = direct_codependencies(codeps, addons_set, result)
        result |= codeps
    return result if include_selected else result - addons_selection


def direct_codependencies(
    root_addons: Set[str], addons_set: AddonsSet, accumulator: Set[str]
) -> Set[str]:
    return {
        a
        for a in addons_set
        if set(addons_set[a].manifest.depends) & root_addons and a not in accumulator
    }



from typing import Iterable, Set, Tuple

from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet
from ..dependency_iterator import dependency_iterator


def list_external_dependencies_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    kind: str,
    transitive: bool = False,
) -> Tuple[Iterable[str], Iterable[str]]:
    result: Set[str] = set()
    missing: Set[str] = set()
    for addon_name, addon in dependency_iterator(
        addons_selection, addons_set, transitive
    ):
        if not addon:
            missing.add(addon_name)
        else:
            result.update(addon.manifest.external_dependencies.get(kind, []))
    return sorted(result), missing



from manifestoo.dependency_iterator import dependency_iterator

from .common import mock_addons_set


def test_basic():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"depends": ["c"]},
        }
    )
    assert sorted(
        addon_name
        for addon_name, _ in dependency_iterator(
            ["a", "c"], addons_set, transitive=False
        )
    ) == ["a", "c"]
    assert sorted(
        addon_name
        for addon_name, _ in dependency_iterator(
            ["a", "c"], addons_set, transitive=True
        )
    ) == ["a", "b", "c"]


def test_loop():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"depends": ["c"]},
            "c": {"depends": ["a"]},
        }
    )
    assert sorted(
        addon_name
        for addon_name, _ in dependency_iterator(
            ["a", "c"], addons_set, transitive=True
        )
    ) == ["a", "b", "c"]



from manifestoo.odoo_series import (
    OdooSeries,
    detect_from_addon_version,
    detect_from_addons_set,
)

from .common import mock_addons_set


def test_detect_from_version():
    assert detect_from_addon_version("1.0.0") is None
    assert detect_from_addon_version("1.0.0.0.0.0") is None
    assert detect_from_addon_version("8.0.1") is None
    assert detect_from_addon_version("8.0.1.0") is None
    assert detect_from_addon_version("8.0.1.0.0") == OdooSeries.v8_0
    assert detect_from_addon_version("14.0.1.0.0.1") == OdooSeries.v14_0


def test_detect_none():
    addons_set = mock_addons_set(
        {
            "a": {"version": "1.0.0"},
            "b": {},
        }
    )
    assert detect_from_addons_set(addons_set) is None


def test_detect_one():
    addons_set = mock_addons_set(
        {
            "a": {"version": "1.0.0"},
            "b": {"version": "12.0.1.0.0"},
        }
    )
    assert detect_from_addons_set(addons_set) is OdooSeries.v12_0


def test_detect_ambiguous():
    addons_set = mock_addons_set(
        {
            "a": {"version": "13.0.1.0.0"},
            "b": {"version": "12.0.1.0.0"},
        }
    )
    assert detect_from_addons_set(addons_set) is None



import contextlib

import pytest

from manifestoo import echo


@contextlib.contextmanager
def _verbosity(n):
    saved_verbosity = echo.verbosity
    echo.verbosity = n
    try:
        yield
    finally:
        echo.verbosity = saved_verbosity


@pytest.mark.parametrize()
def _test_verbosity(func, threshold, capsys):
    with _verbosity(threshold - 1):
        # no output below threshold
        func("msg")
        assert "msg" not in capsys.readouterr().err
    with _verbosity(threshold):
        # verbosity on or above verbosity
        func("msg")
        assert capsys.readouterr().err == "msg\n"


def test_debug(capsys):
    _test_verbosity(echo.debug, 2, capsys)


def test_info(capsys):
    _test_verbosity(echo.info, 1, capsys)


def test_notice(capsys):
    _test_verbosity(echo.notice, 0, capsys)


def test_warning(capsys):
    _test_verbosity(echo.warning, -1, capsys)


def test_echo_error(capsys):
    echo.error("msg")
    assert capsys.readouterr().err == "msg\n"


def test_echo_error_low_verbosity(capsys):
    # no threshold for error messages
    with _verbosity(-10):
        echo.error("msg")
        assert capsys.readouterr().err == "msg\n"


def test_echo_to_stdout(capsys):
    echo.error("msg", err=False)
    assert capsys.readouterr().out == "msg\n"



from pathlib import Path
from typing import List, Set

from .addons_set import AddonsSet
from .utils import comma_split


class AddonsSelection(Set[str]):
    def __str__(self) -> str:
        return ",".join(sorted(self))

    def add_addons_dirs(self, addons_dirs: List[Path]) -> None:
        addons_set = AddonsSet()
        addons_set.add_from_addons_dirs(addons_dirs)
        self.update(addons_set.keys())

    def add_addon_names(self, addon_names: str) -> None:
        for addon_name in comma_split(addon_names):
            self.add(addon_name)

    def remove_addon_names(self, addon_names: str) -> None:
        for addon_name in comma_split(addon_names):
            try:
                self.remove(addon_name)
            except KeyError:
                pass



from pathlib import Path
from typing import Dict, Iterable

from . import echo
from .addon import Addon, AddonNotFound


class AddonsSet(Dict[str, Addon]):
    def __str__(self) -> str:
        return ",".join(sorted(self.keys()))

    def add_from_addons_dir(self, addons_dir: Path) -> None:
        if not addons_dir.is_dir():
            echo.notice(f"ignoring {addons_dir}: not a directory")
            return
        for addon_dir in addons_dir.iterdir():
            try:
                addon = Addon.from_addon_dir(addon_dir)
            except AddonNotFound as e:
                echo.debug(f"ignoring {addon_dir}: {e}")
                continue
            else:
                self[addon.name] = addon

    def add_from_addons_dirs(self, addons_dirs: Iterable[Path]) -> None:
        for addons_dir in addons_dirs:
            self.add_from_addons_dir(addons_dir)



from typer.testing import CliRunner

from manifestoo.commands.list_external_dependencies import (
    list_external_dependencies_command,
)
from manifestoo.main import app

from .common import mock_addons_selection, mock_addons_set, populate_addons_dir


def tests_basic():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"external_dependencies": {"deb": ["imagemagick"]}},
        }
    )
    assert list_external_dependencies_command(
        mock_addons_selection("a"), addons_set, "deb", transitive=False
    ) == (
        [],
        set(),
    )


def tests_transitive():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"external_dependencies": {"deb": ["imagemagick"]}},
        }
    )
    assert list_external_dependencies_command(
        mock_addons_selection("a"), addons_set, "deb", transitive=True
    ) == (
        ["imagemagick"],
        set(),
    )


def tests_missing():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "external_dependencies": {"deb": ["imagemagick"]}},
        }
    )
    assert list_external_dependencies_command(
        mock_addons_selection("a"), addons_set, "deb", True
    ) == (
        ["imagemagick"],
        {"b"},
    )


def test_integration(tmp_path):
    addons = {
        "a": {"depends": ["b"], "external_dependencies": {"deb": ["curl", "wget"]}},
        "b": {"external_dependencies": {"deb": ["imagemagick"]}},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [
            f"--select-addons-dir={tmp_path}",
            "list-external-dependencies",
            "deb",
            "--transitive",
        ],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "curl\nimagemagick\nwget\n"



from typer.testing import CliRunner

from manifestoo.commands.check_dev_status import (
    CORE_DEV_STATUS,
    check_dev_status_command,
)
from manifestoo.main import app
from manifestoo.odoo_series import OdooSeries

from .common import mock_addons_selection, mock_addons_set, populate_addons_dir


def test_missing_dev_status():
    addons_set = mock_addons_set(
        {
            "a": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status=None,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == ["a has missing development_status"]


def test_missing_dev_status_depends():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "development_status": "Stable"},
            "b": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status=None,
        transitive=False,
        odoo_series=OdooSeries.v14_0,
    )
    assert errors == ["b has missing development_status"]


def test_invalid_dev_status():
    addons_set = mock_addons_set(
        {
            "a": {"development_status": "bad"},
            "b": {"development_status": CORE_DEV_STATUS},
        }
    )
    addons_selection = mock_addons_selection("a,b")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status=None,
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert sorted(errors) == [
        "a has invalid development_status 'bad'",
        "b has invalid development_status 'core'",
    ]


def test_invalid_dev_status_depends():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "development_status": "Stable"},
            "b": {"development_status": "bad"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status=None,
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == ["b has invalid development_status 'bad'"]


def test_missing_selection():
    addons_set = mock_addons_set({})
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == ["a not found"]


def test_missing_depend():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == ["b not found"]


def test_ok():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == []


def test_basic():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"development_status": "Alpha"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == ["a (Beta) depends on b (Alpha)"]


def test_double():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "development_status": "Stable"},
            "b": {"depends": ["c"], "development_status": "Beta"},
            "c": {"depends": [], "development_status": "Alpha"},
        }
    )
    addons_selection = mock_addons_selection("a,b")
    errors = check_dev_status_command(
        sorted(addons_selection),
        addons_set,
        default_dev_status="Beta",
        transitive=True,
        odoo_series=OdooSeries.v13_0,
    )
    assert sorted(errors) == [
        "a (Stable) depends on b (Beta)",
        "b (Beta) depends on c (Alpha)",
    ]


def test_transitive():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"depends": ["c"]},
            "c": {"development_status": "Alpha"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=False,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == []
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=True,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == ["b (Beta) depends on c (Alpha)"]


def test_core_addon():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["base"], "development_status": "Stable"},
            "base": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_dev_status_command(
        addons_selection,
        addons_set,
        default_dev_status="Beta",
        transitive=True,
        odoo_series=OdooSeries.v13_0,
    )
    assert errors == []


def test_integration(tmp_path):
    addons = {
        "a": {"depends": ["b"], "development_status": "Beta"},
        "b": {},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [
            f"--addons-path={tmp_path}",
            "--select-include=a",
            "--odoo-series=13.0",
            "check-dev-status",
            "--default-dev-status=Alpha",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert result.stdout == "a (Beta) depends on b (Alpha)\n"



from typing import Iterable, Optional, Set, Tuple

from .. import echo
from ..addon import Addon
from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet
from ..core_addons import get_core_addon_license, is_core_addon
from ..dependency_iterator import dependency_iterator
from ..license import LicenseType, can_depend_on, get_license_type
from ..odoo_series import OdooSeries


def _get_license_type_or_proprietary(
    addon: Addon, odoo_series: OdooSeries
) -> Tuple[Optional[str], LicenseType]:
    addon_license: Optional[str] = None
    if is_core_addon(addon.name, odoo_series):
        addon_license = get_core_addon_license(addon.name, odoo_series)
    else:
        addon_license = addon.manifest.license
    if not addon_license:
        echo.warning(f"No license declared for {addon.name}, assuming Proprietary.")
        return addon_license, LicenseType.PROPRIETARY
    addon_license_type = get_license_type(addon_license)
    if not addon_license_type:
        echo.warning(
            f"Unknown license {addon_license} for {addon.name}, "
            "assuming Proprietary."
        )
        return addon_license, LicenseType.PROPRIETARY
    return addon_license, addon_license_type


def check_licenses_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    transitive: bool,
    odoo_series: OdooSeries,
) -> Iterable[str]:
    errors: Set[str] = set()
    for addon_name, addon in dependency_iterator(
        addons_selection, addons_set, transitive
    ):
        if not addon:
            errors.add(f"{addon_name} not found")
            continue
        addon_license, addon_license_type = _get_license_type_or_proprietary(
            addon, odoo_series
        )
        for depend_name in addon.manifest.depends:
            depend = addons_set.get(depend_name)
            if not depend:
                errors.add(f"{depend_name} not found")
                continue
            depend_license, depend_license_type = _get_license_type_or_proprietary(
                depend, odoo_series
            )
            if not can_depend_on(addon_license_type, depend_license_type):
                errors.add(
                    f"{addon_name} ({addon_license}) depends on "
                    f"{depend_name} ({depend_license})"
                )
    return sorted(errors)



import pytest
import typer

from manifestoo.utils import comma_split, not_implemented, notice_or_abort, print_list


@pytest.mark.parametrize(
    "s, expected",
    [
        (None, []),
        ("", []),
        ("  ", []),
        (" a", ["a"]),
        ("a, b", ["a", "b"]),
        ("a,,b, c ", ["a", "b", "c"]),
    ],
)
def test_comma_split(s, expected):
    assert comma_split(s) == expected


def test_print_list(capsys):
    print_list(["b", "a"], ",")
    assert capsys.readouterr().out == "b,a\n"


def test_print_empty_list(capsys):
    print_list([], ",")
    assert capsys.readouterr().out == ""


def test_not_implemented(capsys):
    with pytest.raises(typer.Abort):
        not_implemented("msg")
    assert capsys.readouterr().err == "msg is not implemented.\n"


def test_notice_or_abort(capsys):
    notice_or_abort("msg", False)
    assert capsys.readouterr().err == "msg\n"
    with pytest.raises(typer.Abort):
        notice_or_abort("msg", True)
    assert capsys.readouterr().err == "msg\n"



from pathlib import Path

from .manifest import InvalidManifest, Manifest


class AddonNotFound(Exception):
    pass


class AddonNotFoundNotInstallable(AddonNotFound):
    pass


class AddonNotFoundNoInit(AddonNotFound):
    pass


class AddonNotFoundNoManifest(AddonNotFound):
    pass


class AddonNotFoundNotADirectory(AddonNotFound):
    pass


class AddonNotFoundInvalidManifest(AddonNotFound):
    pass


def _get_manifest_path(addon_dir: Path) -> Path:
    for manifest_name in ("__manifest__.py", "__openerp__.py", "__terp__.py"):
        manifest_path = addon_dir / manifest_name
        if manifest_path.is_file():
            return manifest_path
    raise AddonNotFoundNoManifest(f"No manifest found in {addon_dir}")


class Addon:
    def __init__(self, manifest: Manifest):
        self.manifest = manifest
        self.manifest_path = manifest.manifest_path
        self.path = self.manifest_path.parent
        self.name = self.path.name

    @classmethod
    def from_addon_dir(
        cls, addon_dir: Path, allow_not_installable: bool = False
    ) -> "Addon":
        if not addon_dir.is_dir():
            raise AddonNotFoundNotADirectory(f"{addon_dir} is not a directory")
        manifest_path = _get_manifest_path(addon_dir)
        try:
            manifest = Manifest.from_manifest_path(manifest_path)
            if not allow_not_installable and not manifest.installable:
                raise AddonNotFoundNotInstallable(f"{addon_dir} is not installable")
        except InvalidManifest as e:
            raise AddonNotFoundInvalidManifest(str(e)) from e
        if not addon_dir.joinpath("__init__.py").is_file():
            raise AddonNotFoundNoInit(f"{addon_dir} is missing an __init__.py")
        return cls(manifest)



from manifestoo.addons_set import AddonsSet

from .common import populate_addons_dir


def test_from_addons_dir(tmp_path):
    addons = {
        "a": {},
        "b": {},
        "c": {"installable": False},
    }
    populate_addons_dir(tmp_path, addons)
    addons_set = AddonsSet()
    addons_set.add_from_addons_dir(tmp_path)
    assert str(addons_set) == "a,b"


def test_from_addons_dirs(tmp_path):
    addons1 = {
        "a": {},
        "b": {},
    }
    addons_dir_1 = tmp_path / "addons_dir_1"
    populate_addons_dir(addons_dir_1, addons1)
    addons2 = {
        "c": {},
    }
    addons_dir_2 = tmp_path / "addons_dir_2"
    populate_addons_dir(addons_dir_2, addons2)
    addons_set = AddonsSet()
    addons_set.add_from_addons_dirs([addons_dir_1, addons_dir_2])
    assert str(addons_set) == "a,b,c"


def test_from_missing_dir(tmp_path):
    addons_set = AddonsSet()
    addons_set.add_from_addons_dirs([tmp_path / "not-a-dir"])
    assert str(addons_set) == ""



import sys
from typing import Iterable, List, Optional

import typer

from . import echo
from .odoo_series import OdooSeries


def comma_split(s: Optional[str]) -> List[str]:
    if not s:
        return []
    s = s.strip()
    if not s:
        return []
    items = [item.strip() for item in s.split(",")]
    return [item for item in items if item]


def not_implemented(what: str) -> None:
    echo.error(f"{what} is not implemented.")
    raise typer.Abort()


def print_list(lst: Iterable[str], separator: str) -> None:
    if not lst:
        return
    sys.stdout.write(separator.join(lst))
    sys.stdout.write("\n")


def notice_or_abort(msg: str, abort: bool) -> None:
    if abort:
        echo.error(msg)
        raise typer.Abort()
    else:
        echo.notice(msg)


def ensure_odoo_series(odoo_series: Optional[OdooSeries]) -> None:
    if not odoo_series:
        echo.error(
            "Odoo series could not be detected. Please provide one with --odoo-series."
        )
        raise typer.Abort()



import textwrap

from typer.testing import CliRunner

from manifestoo.main import app

from .common import populate_addons_dir


def test_integration(tmp_path):
    addons = {
        "a": {"version": "13.0.1.0.0", "depends": ["b", "c"]},
        "b": {"depends": ["base", "mail"]},
        "c": {"depends": ["account", "b"]},
        "account": {"depends": ["base"]},
        "base": {},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        ["--select=a", f"--addons-path={tmp_path}", "tree"],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == textwrap.dedent(
        """\
            a (13.0.1.0.0)
            ├── b (no version)
            │   └── mail (✘ not installed)
            └── c (no version)
                ├── account (13.0+c)
                └── b ⬆
        """
    )



from pathlib import Path
from typing import Any, Dict

from manifestoo.addon import Addon
from manifestoo.addons_selection import AddonsSelection
from manifestoo.addons_set import AddonsSet
from manifestoo.manifest import Manifest


def populate_addons_dir(addons_dir: Path, addons: Dict[str, Dict[str, Any]]):
    if not addons_dir.is_dir():
        addons_dir.mkdir()
    for addon_name, manifest in addons.items():
        addon_path = addons_dir / addon_name
        addon_path.mkdir()
        (addon_path / "__init__.py").touch()
        manifest_path = addon_path / "__manifest__.py"
        manifest_path.write_text(repr(manifest))


def mock_manifest(
    manifest_dict: Dict[str, Any], addon_name: str = "fake-addon"
) -> Manifest:
    manifest_path = Path("/tmp/fake-addons-dir") / addon_name / "__manifest__.py"
    return Manifest(manifest_path, manifest_dict)


def mock_addons_set(addons: Dict[str, Dict[str, Any]]) -> AddonsSet:
    addons_set = AddonsSet()
    for addon_name, manifest_dict in addons.items():
        manifest = mock_manifest(manifest_dict, addon_name)
        addons_set[addon_name] = Addon(manifest)
    return addons_set


def mock_addons_selection(addon_names: str) -> AddonsSelection:
    addons_selection = AddonsSelection()
    addons_selection.add_addon_names(addon_names)
    return addons_selection



from manifestoo.addons_selection import AddonsSelection

from .common import populate_addons_dir


def test_add_addons_dirs(tmp_path):
    addons = {
        "a": {},
        "b": {},
        "c": {"installable": False},
    }
    populate_addons_dir(tmp_path, addons)
    addons_selection = AddonsSelection()
    addons_selection.add_addons_dirs([tmp_path])
    assert str(addons_selection) == "a,b"  # c is not installable


def test_add_addon_names():
    addons_selection = AddonsSelection()
    addons_selection.add_addon_names("a,b")
    addons_selection.add_addon_names("a,c")
    assert str(addons_selection) == "a,b,c"


def test_add_remove_addon_names():
    addons_selection = AddonsSelection()
    addons_selection.add_addon_names("a,b,c,d")
    addons_selection.remove_addon_names("a,c,e")
    assert str(addons_selection) == "b,d"



from typer.testing import CliRunner

from manifestoo.commands.check_licenses import check_licenses_command
from manifestoo.main import app
from manifestoo.odoo_series import OdooSeries

from .common import mock_addons_selection, mock_addons_set, populate_addons_dir


def test_missing_license():
    addons_set = mock_addons_set(
        {
            "a": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == []


def test_mit_depends_mit():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "license": "MIT"},
            "b": {"license": "MIT"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == []


def test_unknown_license_cant_depend_on_open_source():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "license": "My License"},
            "b": {"license": "GPL-3"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == ["a (My License) depends on b (GPL-3)"]


def test_unknown_license_can_depend_on_proprietary():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"], "license": "My License"},
            "b": {"license": "Other Proprietary"},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == []


def test_v8_agpl():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["account"], "license": "LGPL-3"},
            "account": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v8_0,
    )
    assert errors == ["a (LGPL-3) depends on account (AGPL-3)"]


def test_v12_ce_lgpl():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["account"], "license": "LGPL-3"},
            "account": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == []


def test_v12_ee_proprietary():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["account_accountant"], "license": "LGPL-3"},
            "account_accountant": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    errors = check_licenses_command(
        addons_selection,
        addons_set,
        transitive=False,
        odoo_series=OdooSeries.v12_0,
    )
    assert errors == ["a (LGPL-3) depends on account_accountant (OEEL-1)"]


def test_integration(tmp_path):
    addons = {
        "a": {"depends": ["b"], "license": "GPL-3"},
        "b": {"depends": ["c", "d"], "license": "AGPL-3"},
        "d": {"license": "LGPL-3"},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [
            f"--addons-path={tmp_path}",
            "--select-include=a",
            "--odoo-series=13.0",
            "check-licenses",
            "--transitive",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert result.stdout == "a (GPL-3) depends on b (AGPL-3)\nc not found\n"



from typer.testing import CliRunner

from manifestoo.main import app

from .common import populate_addons_dir


def test_list_missing(tmp_path):
    addons = {
        "a": {"depends": ["c"]},
        "b": {"depends": ["d"]},
        "e": {"depends": ["a", "b"]},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [f"--select-addons-dir={tmp_path}", "list-missing"],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "c\nd\n"



from enum import Enum
from typing import Dict, Optional


class LicenseType(Enum):
    PROPRIETARY = 1
    PERMISSIVE = 2  # MIT, Apache
    WEAKLY_PROTECTIVE = 3  # LGPL
    STRONGLY_PROTECTIVE = 4  # GPL
    NETWORK_PROTECTIVE = 5  # AGPL


def can_depend_on(work_license: LicenseType, dependency_license: LicenseType) -> bool:
    if work_license == LicenseType.PROPRIETARY:
        return dependency_license in (
            LicenseType.PROPRIETARY,
            LicenseType.PERMISSIVE,
            LicenseType.WEAKLY_PROTECTIVE,
        )
    elif work_license == LicenseType.PERMISSIVE:
        return dependency_license in (LicenseType.PERMISSIVE,)
    elif work_license == LicenseType.WEAKLY_PROTECTIVE:
        return dependency_license in (
            LicenseType.WEAKLY_PROTECTIVE,
            LicenseType.PERMISSIVE,
        )
    elif work_license == LicenseType.STRONGLY_PROTECTIVE:
        return dependency_license in (
            LicenseType.STRONGLY_PROTECTIVE,
            LicenseType.WEAKLY_PROTECTIVE,
            LicenseType.PERMISSIVE,
        )
    elif work_license == LicenseType.NETWORK_PROTECTIVE:
        return dependency_license in (
            LicenseType.NETWORK_PROTECTIVE,
            LicenseType.STRONGLY_PROTECTIVE,
            LicenseType.WEAKLY_PROTECTIVE,
            LicenseType.PERMISSIVE,
        )
    raise AssertionError(
        f"Unhandled license type {work_license!r}."
    )  # pragma: no cover


_licenses: Dict[str, LicenseType] = {
    "other proprietary": LicenseType.PROPRIETARY,
    "oeel-1": LicenseType.PROPRIETARY,
    "opl-1": LicenseType.PROPRIETARY,
    "agpl-3": LicenseType.NETWORK_PROTECTIVE,
    "gpl-3": LicenseType.STRONGLY_PROTECTIVE,
    "gpl-3 or any later version": LicenseType.STRONGLY_PROTECTIVE,
    "lgpl-3": LicenseType.WEAKLY_PROTECTIVE,
    "lgpl-3 or any later version": LicenseType.WEAKLY_PROTECTIVE,
    "mit": LicenseType.PERMISSIVE,
}


def get_license_type(license: str) -> Optional[LicenseType]:
    return _licenses.get(license.lower())



from typing import Iterable, Iterator, Optional, Set, Tuple

from .addon import Addon
from .addons_selection import AddonsSelection
from .addons_set import AddonsSet


def dependency_iterator(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    transitive: bool,
) -> Iterator[Tuple[str, Optional[Addon]]]:
    """Iterate addons and their dependencies.

    Yield tuples:
    - addon name
    - addon object (None if not found in addons_set)

    If transitive is False, only yield addon_selection.

    An addon is yielded at most once.
    """
    done: Set[str] = set()

    def _iter(
        addon_names: Iterable[str],
    ) -> Iterator[Tuple[str, Optional[Addon]]]:
        done.update(addon_names)
        for addon_name in addon_names:
            addon = addons_set.get(addon_name)
            yield addon_name, addon
            if transitive and addon:
                yield from _iter(set(addon.manifest.depends) - done)

    yield from _iter(addons_selection)



from typer.testing import CliRunner

from manifestoo.commands.list_depends import list_depends_command
from manifestoo.main import app

from .common import mock_addons_selection, mock_addons_set, populate_addons_dir


def test_basic():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    assert list_depends_command(addons_selection, addons_set, transitive=False) == (
        ["b"],
        set(),
    )


def test_transitive():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"depends": ["c"]},
            "c": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    assert list_depends_command(addons_selection, addons_set, transitive=False) == (
        ["b"],
        set(),
    )
    assert list_depends_command(addons_selection, addons_set, transitive=True) == (
        [
            "b",
            "c",
        ],
        set(),
    )


def test_loop():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {"depends": ["c"]},
            "c": {"depends": ["a"]},
        }
    )
    addons_selection = mock_addons_selection("a")
    assert list_depends_command(addons_selection, addons_set, transitive=True) == (
        [
            "b",
            "c",
        ],
        set(),
    )
    assert list_depends_command(
        addons_selection, addons_set, include_selected=True, transitive=True
    ) == (
        [
            "a",
            "b",
            "c",
        ],
        set(),
    )


def test_missing():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
        }
    )
    assert list_depends_command(
        mock_addons_selection("a"), addons_set, transitive=False
    ) == (
        [
            "b",
        ],
        set(),
    )
    assert list_depends_command(
        mock_addons_selection("a"), addons_set, transitive=True
    ) == (
        [
            "b",
        ],
        {"b"},
    )
    assert list_depends_command(
        mock_addons_selection("a,c"), addons_set, transitive=True
    ) == (
        [
            "b",
        ],
        {"b", "c"},
    )
    assert list_depends_command(
        mock_addons_selection("a,c"), addons_set, include_selected=True, transitive=True
    ) == (
        ["a", "b", "c"],
        {"b", "c"},
    )


def test_include_selected_not_included():
    """Dependencies that are part of the selection are not returned."""
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {},
        }
    )
    addons_selection = mock_addons_selection("a,b")
    assert list_depends_command(
        addons_selection, addons_set, include_selected=False
    ) == (
        [],
        set(),
    )


def test_include_selected():
    addons_set = mock_addons_set(
        {
            "a": {"depends": ["b"]},
            "b": {},
        }
    )
    addons_selection = mock_addons_selection("a")
    assert list_depends_command(
        addons_selection, addons_set, include_selected=True
    ) == (
        [
            "a",
            "b",
        ],
        set(),
    )


def test_integration(tmp_path):
    addons = {
        "a": {"depends": ["b"]},
        "b": {},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [f"--addons-path={tmp_path}", "--select-include", "a", "list-depends"],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "b\n"



import sys
from pathlib import Path

from manifestoo.addons_path import AddonsPath


def test_from_addons_dirs():
    addons_path = AddonsPath()
    addons_path.extend_from_addons_dirs([Path("a"), Path("b")])
    assert str(addons_path) == "a,b"


def test_from_addons_path():
    addons_path = AddonsPath()
    addons_path.extend_from_addons_path("a,b")
    assert str(addons_path) == "a,b"


def test_from_odoo_cfg(tmp_path):
    odoo_cfg = tmp_path / "odoo.cfg"
    odoo_cfg.write_text("[options]\naddons_path=a,b\n")
    addons_path = AddonsPath()
    addons_path.extend_from_odoo_cfg(odoo_cfg)
    assert str(addons_path) == "a,b"


def test_from_odoo_cfg__without_addons_path(tmp_path):
    odoo_cfg = tmp_path / "odoo.cfg"
    odoo_cfg.write_text("[options]\n")
    addons_path = AddonsPath()
    addons_path.extend_from_odoo_cfg(odoo_cfg)
    assert str(addons_path) == ""


def test_from_import_odoo__python_not_found():
    addons_path = AddonsPath()
    addons_path.extend_from_import_odoo("this-is-not-a-snake")
    assert str(addons_path) == ""


def test_from_import_odoo__odoo_not_found():
    addons_path = AddonsPath()
    addons_path.extend_from_import_odoo(sys.executable)
    assert str(addons_path) == ""



from typing import Optional

import typer

verbosity = 0


def debug(msg: str) -> None:
    if verbosity > 1:
        typer.secho(msg, dim=True, err=True)


def info(
    msg: str, nl: bool = True, bold: bool = False, bold_intro: Optional[str] = None
) -> None:
    if verbosity > 0:
        if bold_intro:
            typer.secho(
                bold_intro, fg=typer.colors.BRIGHT_BLUE, err=True, nl=False, bold=True
            )
        typer.secho(msg, fg=typer.colors.BRIGHT_BLUE, err=True, nl=nl, bold=bold)


def notice(msg: str, nl: bool = True, bold: bool = False) -> None:
    if verbosity > -1:
        typer.secho(msg, fg=typer.colors.GREEN, err=True, nl=nl, bold=bold)


def warning(msg: str) -> None:
    if verbosity > -2:
        typer.secho(msg, fg=typer.colors.YELLOW, err=True)


def error(msg: str, err: bool = True) -> None:
    typer.secho(msg, fg=typer.colors.RED, err=err)



from enum import Enum
from typing import Optional, Set

from . import echo
from .addons_set import AddonsSet


class OdooSeries(str, Enum):
    v8_0 = "8.0"
    v9_0 = "9.0"
    v10_0 = "10.0"
    v11_0 = "11.0"
    v12_0 = "12.0"
    v13_0 = "13.0"
    v14_0 = "14.0"
    v15_0 = "15.0"
    v16_0 = "16.0"


class OdooEdition(str, Enum):
    CE = "c"
    EE = "e"


def detect_from_addon_version(version: str) -> Optional[OdooSeries]:
    parts = version.split(".")
    if len(parts) < 5:
        return None
    try:
        return OdooSeries["v" + "_".join(parts[:2])]
    except KeyError:
        return None


def detect_from_addons_set(addons_set: AddonsSet) -> Optional[OdooSeries]:
    detected: Set[OdooSeries] = set()
    for addon in addons_set.values():
        addon_version = addon.manifest.version
        if not addon_version:
            continue
        addon_series = detect_from_addon_version(addon_version)
        if not addon_series:
            continue
        detected.add(addon_series)
    if len(detected) == 0:
        echo.notice("No Odoo series detected in addons set")
        return None
    elif len(detected) > 1:
        echo.notice(
            f"Different Odoo series detected in addons set: {', '.join(detected)}"
        )
        return None
    return detected.pop()



from pathlib import Path

import pytest

from manifestoo.addon import (
    Addon,
    AddonNotFoundInvalidManifest,
    AddonNotFoundNoInit,
    AddonNotFoundNoManifest,
    AddonNotFoundNotADirectory,
    AddonNotFoundNotInstallable,
)


@pytest.fixture(
    params=[
        {
            "dir": "a",
            "manifest": "{}",
        }
    ]
)
def addon_dir(request, tmp_path) -> Path:
    addon_dir = tmp_path / request.param["dir"]
    addon_dir.mkdir()
    (addon_dir / "__init__.py").touch()
    (addon_dir / "__manifest__.py").write_text(request.param["manifest"])
    return addon_dir


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "theaddon",
            "manifest": "{'name': 'the addon'}",
        }
    ],
    indirect=True,
)
def test_basic(addon_dir):
    addon = Addon.from_addon_dir(addon_dir)
    assert addon.name == "theaddon"
    assert addon.path == addon_dir
    assert addon.manifest.name == "the addon"


def test_not_a_directory(tmp_path):
    with pytest.raises(AddonNotFoundNotADirectory):
        Addon.from_addon_dir(tmp_path / "not-a-dir")


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable': False}",
        }
    ],
    indirect=True,
)
def test_not_installable(addon_dir):
    with pytest.raises(AddonNotFoundNotInstallable):
        Addon.from_addon_dir(addon_dir)
    assert Addon.from_addon_dir(addon_dir, allow_not_installable=True)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "[]",
        }
    ],
    indirect=True,
)
def test_invalid_manifest(addon_dir):
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable':}",
        }
    ],
    indirect=True,
)
def test_manifest_syntax_error(addon_dir):
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable': '?'}",
        }
    ],
    indirect=True,
)
def test_manifest_type_error(addon_dir):
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


def test_no_manifest(addon_dir):
    (addon_dir / "__manifest__.py").unlink()
    with pytest.raises(AddonNotFoundNoManifest):
        Addon.from_addon_dir(addon_dir)


def test_no_init(addon_dir):
    (addon_dir / "__init__.py").unlink()
    with pytest.raises(AddonNotFoundNoInit):
        Addon.from_addon_dir(addon_dir)



import pytest

from manifestoo.manifest import InvalidManifest

from .common import mock_manifest


@pytest.mark.parametrize(
    "key, value",
    [
        ("name", "the name"),
        ("name", None),
        ("version", "1.0.0"),
        ("version", None),
        ("license", "GPL-3"),
        ("license", None),
        ("development_status", "Beta"),
        ("development_status", None),
        ("depends", ["a", "b"]),
        ("external_dependencies", {"python": ["httpx"]}),
        ("installable", True),
        ("installable", False),
    ],
)
def test_manifest_valid_value(key, value):
    manifest = mock_manifest({key: value})
    assert getattr(manifest, key) == value


@pytest.mark.parametrize(
    "key, value",
    [
        ("name", 1),
        ("version", 1),
        ("license", ["i"]),
        ("development_status", {}),
        ("depends", 1),
        ("depends", {}),
        ("depends", None),
        ("external_dependencies", {1: {}}),
        ("external_dependencies", None),
        ("external_dependencies", [1]),
        ("installable", 1),
        ("installable", None),
    ],
)
def test_manifest_invalid_value(key, value):
    manifest = mock_manifest({key: value})
    with pytest.raises(InvalidManifest):
        getattr(manifest, key)


@pytest.mark.parametrize(
    "key, default",
    [
        ("name", None),
        ("version", None),
        ("license", None),
        ("development_status", None),
        ("depends", []),
        ("external_dependencies", {}),
        ("installable", True),
    ],
)
def test_manifest_default_value(key, default):
    manifest = mock_manifest({})
    assert getattr(manifest, key) == default



from typer.testing import CliRunner

from manifestoo.commands.list import list_command
from manifestoo.main import app

from .common import mock_addons_selection, populate_addons_dir


def test_basic():
    addons_selection = mock_addons_selection("b,a")
    assert list_command(addons_selection) == ["a", "b"]


def test_integration(tmp_path):
    addons = {
        "a": {},
        "b": {},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        [f"--select-addons-dir={tmp_path}", "list"],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "a\nb\n"



from typing import Dict, List, Optional, Set

import typer

from ..addon import Addon
from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet
from ..core_addons import is_core_addon, is_core_ce_addon, is_core_ee_addon
from ..odoo_series import OdooEdition, OdooSeries

NodeKey = str


class Node:
    def __init__(self, addon_name: str, addon: Optional[Addon]):
        self.addon_name = addon_name
        self.addon = addon
        self.children = []  # type: List[Node]

    @staticmethod
    def key(addon_name: str) -> NodeKey:
        return addon_name

    def print(self, odoo_series: OdooSeries, fold_core_addons: bool) -> None:
        seen = set()  # type: Set[Node]

        def _print(indent: List[str], node: Node) -> None:
            # inspired by https://stackoverflow.com/a/59109706
            SPACE = "    "
            BRANCH = "│   "
            TEE = "├── "
            LAST = "└── "
            typer.echo(f"{''.join(indent)}{node.addon_name}", nl=False)
            if node in seen:
                typer.secho(" ⬆", dim=True)
                return
            typer.secho(f" ({node.sversion(odoo_series)})", dim=True)
            seen.add(node)
            if not node.children:
                return
            if fold_core_addons and is_core_addon(node.addon_name, odoo_series):
                return
            pointers = [TEE] * (len(node.children) - 1) + [LAST]
            for pointer, child in zip(
                pointers, sorted(node.children, key=lambda n: n.addon_name)
            ):
                if indent:
                    if indent[-1] == TEE:
                        _print(indent[:-1] + [BRANCH, pointer], child)
                    else:
                        assert indent[-1] == LAST
                        _print(indent[:-1] + [SPACE, pointer], child)
                else:
                    _print([pointer], child)

        _print([], self)

    def sversion(self, odoo_series: OdooSeries) -> str:
        if not self.addon:
            return typer.style("✘ not installed", fg=typer.colors.RED)  # type: ignore
        elif is_core_ce_addon(self.addon_name, odoo_series):
            return f"{odoo_series}+{OdooEdition.CE}"
        elif is_core_ee_addon(self.addon_name, odoo_series):
            return f"{odoo_series}+{OdooEdition.EE}"
        else:
            return self.addon.manifest.version or "no version"


def tree_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    odoo_series: OdooSeries,
    fold_core_addons: bool,
) -> None:
    nodes: Dict[NodeKey, Node] = {}

    def add(addon_name: str) -> Node:
        key = Node.key(addon_name)
        if key in nodes:
            return nodes[key]
        addon = addons_set.get(addon_name)
        node = Node(addon_name, addon)
        nodes[key] = node
        if not addon:
            # not found
            return node
        for depend in addon.manifest.depends:
            if depend == "base":
                continue
            node.children.append(add(depend))
        return node

    root_nodes: List[Node] = []
    for addon_name in sorted(addons_selection):
        if addon_name == "base":
            continue
        root_nodes.append(add(addon_name))
    for root_node in root_nodes:
        root_node.print(odoo_series, fold_core_addons)



from typing import Optional

from .addons_path import AddonsPath
from .addons_selection import AddonsSelection
from .addons_set import AddonsSet
from .odoo_series import OdooSeries


class MainOptions:
    def __init__(self) -> None:
        self.addons_path = AddonsPath()
        self.addons_set = AddonsSet()
        self.addons_selection = AddonsSelection()
        self.separator: Optional[str] = None  # deprecated
        self.odoo_series: Optional[OdooSeries] = None



import ast
import os
import subprocess
import tempfile
from configparser import ConfigParser
from pathlib import Path
from typing import Iterable, List

from . import echo
from .utils import comma_split

ADDONS_PATH_SCRIPT = b"""\
import sys
try:
    import openerp as odoo
except ImportError:
    import odoo

odoo.modules.initialize_sys_path()

if odoo.release.version_info >= (13, ):
    path = odoo.addons.__path__
else:
    path = odoo.modules.module.ad_paths

with open(sys.argv[1], "wb") as f:
    f.write(repr(path).encode("utf-8"))
"""


class AddonsPath(List[Path]):
    def __str__(self) -> str:
        return ",".join(str(item) for item in self)

    def extend_from_addons_dirs(self, addons_dirs: Iterable[Path]) -> None:
        self.extend(addons_dirs)

    def extend_from_addons_path(self, addons_path: str) -> None:
        return self.extend(Path(item) for item in comma_split(addons_path))

    def extend_from_odoo_cfg(self, odoo_cfg_path: Path) -> None:
        config = ConfigParser()
        config.read(odoo_cfg_path)
        addons_path = config.get("options", "addons_path", fallback=None)
        if not addons_path:
            return None
        self.extend_from_addons_path(addons_path)

    def extend_from_import_odoo(self, python: str) -> None:
        script = tempfile.NamedTemporaryFile(delete=False, suffix=".py")
        try:
            script.write(ADDONS_PATH_SCRIPT)
            script.close()
            output = tempfile.NamedTemporaryFile(delete=False)
            try:
                output.close()
                try:
                    r = subprocess.call(
                        [python, script.name, output.name],
                        stderr=subprocess.DEVNULL,
                        stdout=subprocess.DEVNULL,
                        env=os.environ,
                    )
                except FileNotFoundError:
                    # python not found
                    r = 1
                if r != 0:
                    echo.notice(f"could not obtain odoo.addons.__path__ using {python}")
                    return
                addons_paths = ast.literal_eval(Path(output.name).read_text())
                self.extend(Path(item) for item in addons_paths)
            finally:
                os.unlink(output.name)
        finally:
            os.unlink(script.name)



from typing import Iterable, Optional, Set

from ..addon import Addon
from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet
from ..core_addons import is_core_addon
from ..dependency_iterator import dependency_iterator
from ..odoo_series import OdooSeries

CORE_DEV_STATUS = "core"
CORE_DEV_STATUS_LEVEL = 100
DEV_STATUS_LEVELS = {
    "alpha": 1,
    "beta": 2,
    "production/stable": 3,
    "production": 3,
    "stable": 3,
    "mature": 4,
}


def _get_dev_status(
    addon: Addon,
    default_dev_status: Optional[str],
    odoo_series: OdooSeries,
    errors: Set[str],
) -> Optional[str]:
    if is_core_addon(addon.name, odoo_series):
        return CORE_DEV_STATUS
    dev_status = addon.manifest.development_status or default_dev_status
    if not dev_status:
        errors.add(f"{addon.name} has missing development_status")
        return None
    if dev_status.lower() not in DEV_STATUS_LEVELS:
        errors.add(f"{addon.name} has invalid development_status {dev_status!r}")
        return None
    return dev_status


def _get_dev_status_level(dev_status: str) -> int:
    if dev_status.lower() == CORE_DEV_STATUS:
        return CORE_DEV_STATUS_LEVEL
    return DEV_STATUS_LEVELS[dev_status.lower()]


def check_dev_status_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    default_dev_status: Optional[str],
    transitive: bool,
    odoo_series: OdooSeries,
) -> Iterable[str]:
    errors: Set[str] = set()
    for addon_name, addon in dependency_iterator(
        addons_selection, addons_set, transitive
    ):
        if not addon:
            errors.add(f"{addon_name} not found")
            continue
        addon_dev_status = _get_dev_status(
            addon, default_dev_status, odoo_series, errors
        )
        if not addon_dev_status:
            continue
        for depend_name in addon.manifest.depends:
            depend = addons_set.get(depend_name)
            if not depend:
                errors.add(f"{depend_name} not found")
                continue
            depend_dev_status = _get_dev_status(
                depend, default_dev_status, odoo_series, errors
            )
            if not depend_dev_status:
                continue
            addon_level = _get_dev_status_level(addon_dev_status)
            depend_level = _get_dev_status_level(depend_dev_status)
            if addon_level > depend_level:
                errors.add(
                    f"{addon_name} ({addon_dev_status}) depends on "
                    f"{depend_name} ({depend_dev_status})"
                )
    return sorted(errors)



from pathlib import Path
from typing import List, Optional

import typer

from . import echo
from .commands.check_dev_status import check_dev_status_command
from .commands.check_licenses import check_licenses_command
from .commands.list import list_command
from .commands.list_codepends import list_codepends_command
from .commands.list_depends import list_depends_command
from .commands.list_external_dependencies import list_external_dependencies_command
from .commands.tree import tree_command
from .core_addons import get_core_addons
from .odoo_series import OdooSeries, detect_from_addons_set
from .options import MainOptions
from .utils import ensure_odoo_series, not_implemented, print_list
from .version import version

app = typer.Typer()


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"manifestoo version {version}")
        raise typer.Exit()


@app.callback()
def callback(
    ctx: typer.Context,
    select_addons_dirs: Optional[List[Path]] = typer.Option(
        None,
        "--select-addons-dir",
        "-d",
        exists=True,
        file_okay=False,
        dir_okay=True,
        help=(
            "Select all installable addons found in this directory. "
            "This option may be repeated. "
            "The directories selected with this options are "
            "automatically added to the addons search path."
        ),
        show_default=False,
    ),
    select_include: Optional[str] = typer.Option(
        None,
        "--select-include",
        "--select",
        metavar="addon1,addon2,...",
        help=(
            "Comma separated list of addons to select. "
            "These addons will be searched in the addons path."
        ),
    ),
    select_exclude: Optional[str] = typer.Option(
        None,
        metavar="addon1,addon2,...",
        help=(
            "Comma separated list of addons to exclude from selection. "
            "This option is useful in combination with `--select-addons-dir`."
        ),
    ),
    select_core_addons: bool = typer.Option(
        False,
        "--select-core-addons",
        help="Select the Odoo core addons (CE and EE) for the given series.",
        show_default=False,
    ),
    addons_path: Optional[str] = typer.Option(
        None,
        help="Expand addons path with this comma separated list of directories.",
    ),
    addons_path_from_import_odoo: bool = typer.Option(
        True,
        help=(
            "Expand addons path by trying to `import odoo` and "
            "looking at `odoo.addons.__path__`. This option is useful when "
            "addons have been installed with pip."
        ),
    ),
    addons_path_python: str = typer.Option(
        "python",
        "--addons-path-python",
        show_default=False,
        metavar="PYTHON",
        help=(
            "The python executable to use when importing `odoo.addons.__path__`. "
            "Defaults to the `python` executable found in PATH."
        ),
    ),
    addons_path_from_odoo_cfg: Optional[Path] = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        envvar="ODOO_RC",
        help=(
            "Expand addons path by looking into the provided Odoo configuration file. "
        ),
    ),
    separator: str = typer.Option(
        default=None,
        hidden=True,  # deprecated
    ),
    odoo_series: Optional[OdooSeries] = typer.Option(
        None,
        envvar=["ODOO_VERSION", "ODOO_SERIES"],
        help="Odoo series to use, in case it is not autodetected from addons version.",
    ),
    verbose: int = typer.Option(
        0,
        "--verbose",
        "-v",
        count=True,
        show_default=False,
    ),
    quiet: int = typer.Option(
        0,
        "--quiet",
        "-q",
        count=True,
        show_default=False,
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
    ),
) -> None:
    """Reason about Odoo addons manifests.

    The `--select-*` options of this command select addons on which the
    subcommands will act. The `--addons-path` options provide locations to
    search for addons.

    Run `manifestoo <subcommand> --help` for more options.
    """
    echo.verbosity += verbose
    echo.verbosity -= quiet
    main_options = MainOptions()
    if separator:
        echo.warning(
            "--separator is deprecated as a global option. "
            "Please use the same option of list, list-depends."
        )
        main_options.separator = separator
    # resolve addons_path
    if select_addons_dirs:
        main_options.addons_path.extend_from_addons_dirs(select_addons_dirs)
    if addons_path:
        main_options.addons_path.extend_from_addons_path(addons_path)
    if addons_path_from_import_odoo:
        main_options.addons_path.extend_from_import_odoo(addons_path_python)
    if addons_path_from_odoo_cfg:
        main_options.addons_path.extend_from_odoo_cfg(addons_path_from_odoo_cfg)
    echo.info(str(main_options.addons_path), bold_intro="Addons path: ")
    # populate addons_set
    main_options.addons_set.add_from_addons_dirs(main_options.addons_path)
    echo.info(str(main_options.addons_set), bold_intro="Addons set: ")
    # Odoo series
    if odoo_series:
        main_options.odoo_series = odoo_series
    else:
        main_options.odoo_series = detect_from_addons_set(main_options.addons_set)
    # addons selection
    if select_addons_dirs:
        main_options.addons_selection.add_addons_dirs(select_addons_dirs)
    if select_include:
        main_options.addons_selection.add_addon_names(select_include)
    if select_exclude:
        main_options.addons_selection.remove_addon_names(select_exclude)
    if select_core_addons:
        ensure_odoo_series(main_options.odoo_series)
        assert main_options.odoo_series
        main_options.addons_selection.update(get_core_addons(main_options.odoo_series))
    if main_options.addons_selection:
        echo.info(str(main_options.addons_selection), bold_intro="Addons selection: ")
    else:
        echo.notice("No addon selected, please use one of the --select options.")
    echo.info(f"{main_options.odoo_series}", bold_intro="Odoo series: ")
    # pass main options to commands
    ctx.obj = main_options


@app.command()
def list(
    ctx: typer.Context,
    separator: Optional[str] = typer.Option(
        None,
        help="Separator character to use (by default, print one item per line).",
    ),
) -> None:
    """Print the selected addons."""
    main_options: MainOptions = ctx.obj
    result = list_command(main_options.addons_selection)
    print_list(result, separator or main_options.separator or "\n")


@app.command()
def list_depends(
    ctx: typer.Context,
    separator: Optional[str] = typer.Option(
        None,
        help="Separator character to use (by default, print one item per line).",
    ),
    transitive: bool = typer.Option(
        False,
        "--transitive",
        help="Print all transitive dependencies.",
        show_default=False,
    ),
    include_selected: bool = typer.Option(
        False,
        "--include-selected",
        help="Print the selected addons along with their dependencies.",
        show_default=False,
    ),
    ignore_missing: bool = typer.Option(
        False,
        "--ignore-missing",
        help=(
            "Do not fail if dependencies are not found in addons path. "
            "This only applies to top level (selected) addons "
            "and transitive dependencies."
        ),
        show_default=False,
    ),
    as_pip_requirements: bool = typer.Option(
        False,
        "--as-pip-requirements",
        show_default=False,
    ),
) -> None:
    """Print the dependencies of selected addons."""
    main_options: MainOptions = ctx.obj
    if as_pip_requirements:
        not_implemented("--as-pip-requirement")
    result, missing = list_depends_command(
        main_options.addons_selection,
        main_options.addons_set,
        transitive,
        include_selected,
    )
    if missing and not ignore_missing:
        echo.error("not found in addons path: " + ",".join(sorted(missing)))
        raise typer.Abort()
    print_list(
        result,
        separator or main_options.separator or "\n",
    )


@app.command()
def list_codepends(
    ctx: typer.Context,
    separator: Optional[str] = typer.Option(
        None,
        help="Separator character to use (by default, print one item per line).",
    ),
    transitive: bool = typer.Option(
        True,
        help="Print all transitive co-dependencies.",
    ),
    include_selected: bool = typer.Option(
        True,
        help="Print the selected addons along with their co-dependencies.",
    ),
) -> None:
    """Print the co-dependencies of selected addons.

    Co-dependencies is the set of addons that depend on the selected
    addons.
    """
    main_options: MainOptions = ctx.obj
    result = list_codepends_command(
        main_options.addons_selection,
        main_options.addons_set,
        transitive,
        include_selected,
    )
    print_list(result, separator or main_options.separator or "\n")


@app.command()
def list_external_dependencies(
    ctx: typer.Context,
    kind: str = typer.Argument(
        ...,
        help="Kind of external dependency, such as `python` or `deb`.",
    ),
    separator: Optional[str] = typer.Option(
        None,
        help="Separator character to use (by default, print one item per line).",
    ),
    transitive: bool = typer.Option(
        False,
        "--transitive",
        help="Print external dependencies of all transitive dependent addons.",
        show_default=False,
    ),
    ignore_missing: bool = typer.Option(
        False,
        "--ignore-missing",
        help=(
            "Do not fail if dependencies are not found in addons path. "
            "This only applies to top level (selected) addons "
            "and transitive dependencies."
        ),
        show_default=False,
    ),
) -> None:
    """Print the external dependencies of selected addons."""
    main_options: MainOptions = ctx.obj
    result, missing = list_external_dependencies_command(
        main_options.addons_selection,
        main_options.addons_set,
        kind,
        transitive,
    )
    if missing and not ignore_missing:
        echo.error("not found in addons path: " + ",".join(sorted(missing)))
        raise typer.Abort()
    print_list(
        result,
        separator or main_options.separator or "\n",
    )


@app.command()
def list_missing(
    ctx: typer.Context,
    separator: Optional[str] = typer.Option(
        None,
        help="Separator character to use (by default, print one item per line).",
    ),
) -> None:
    """Print the missing dependencies of selected addons."""
    main_options: MainOptions = ctx.obj
    result, missing = list_depends_command(
        main_options.addons_selection,
        main_options.addons_set,
        transitive=True,
        include_selected=True,
    )
    print_list(
        sorted(missing),
        separator or main_options.separator or "\n",
    )


@app.command()
def check_licenses(
    ctx: typer.Context,
    transitive: bool = typer.Option(
        False,
        "--transitive",
        help="Also check transitive dependencies.",
        show_default=False,
    ),
) -> None:
    """Check license compatibility.

    Check that selected addons only depend on addons with compatible
    licenses.
    """
    main_options: MainOptions = ctx.obj
    ensure_odoo_series(main_options.odoo_series)
    assert main_options.odoo_series
    errors = check_licenses_command(
        main_options.addons_selection,
        main_options.addons_set,
        transitive,
        main_options.odoo_series,
    )
    if errors:
        echo.error("\n".join(errors), err=False)
        raise typer.Exit(1)


@app.command()
def check_dev_status(
    ctx: typer.Context,
    transitive: bool = typer.Option(
        False,
        "--transitive",
        help="Also check transitive dependencies.",
        show_default=False,
    ),
    default_dev_status: Optional[str] = None,
) -> None:
    """Check development status compatibility.

    Check that selected addons only depend on addons that have an equal
    or higher development status.
    """
    main_options: MainOptions = ctx.obj
    ensure_odoo_series(main_options.odoo_series)
    assert main_options.odoo_series
    errors = check_dev_status_command(
        main_options.addons_selection,
        main_options.addons_set,
        default_dev_status,
        transitive,
        main_options.odoo_series,
    )
    if errors:
        echo.error("\n".join(errors), err=False)
        raise typer.Exit(1)


@app.command()
def tree(
    ctx: typer.Context,
    fold_core_addons: bool = typer.Option(
        False,
        "--fold-core-addons",
        help="Do not expand dependencies of core Odoo addons.",
        show_default=False,
    ),
) -> None:
    """Print the dependency tree of selected addons."""
    main_options: MainOptions = ctx.obj
    ensure_odoo_series(main_options.odoo_series)
    assert main_options.odoo_series
    tree_command(
        main_options.addons_selection,
        main_options.addons_set,
        main_options.odoo_series,
        fold_core_addons,
    )



from typing import Iterable, Set, Tuple

from ..addons_selection import AddonsSelection
from ..addons_set import AddonsSet
from ..dependency_iterator import dependency_iterator


def list_depends_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    transitive: bool = False,
    include_selected: bool = False,
) -> Tuple[Iterable[str], Iterable[str]]:
    result: Set[str] = set()
    missing: Set[str] = set()
    for addon_name, addon in dependency_iterator(
        addons_selection, addons_set, transitive
    ):
        if include_selected:
            result.add(addon_name)
        if not addon:
            missing.add(addon_name)
        else:
            result.update(set(addon.manifest.depends) - set(addons_selection))
    return sorted(result), missing
